/* GStreamer Recorder
 * (c) 2003 Ronald Bultje <rbultje@ronald.bitfreak.net>
 *
 * window.c: user interface setup
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <string.h>
#include <glib.h>
#include <gst/gst.h>
#include <gnome.h>
#include <libgnome/libgnome.h>
#include <libgnomeui/libgnomeui.h>
#include <libgnomevfs/gnome-vfs.h>
#include <gconf/gconf-client.h>

#include <gst/mixer/mixer.h>
#include <gst/tuner/tuner.h>

#include "debug.h"
#include "druid.h"
#include "keys.h"
#include "list.h"
#include "preferences.h"
#include "record.h"
#include "sliders.h"
#include "stats.h"
#include "stock.h"
#include "tv.h"
#include "window.h"

/*
 * Collection of several objects - useful.
 */

#define GST_REC_COLLECTION(a) ((GstRecCollection *) a)
typedef struct _GstRecCollection
{
  GstRec *rec;
  GnomeApp *window;
  GstRecTv *tv;
  GstRecStats *stats;
  GnomeVFSURI *file;
  GnomeFileEntry *fileentry;
  GstElement *output;
  GConfClient *client;
}
GstRecCollection;

/*
 * Popup error.
 */

#define gst_rec_popup_error(coll, msg, args...) \
  do { \
    GtkWidget *window; \
    window = gtk_message_dialog_new (GTK_WINDOW (coll->window), \
				     GTK_DIALOG_MODAL, \
				     GTK_MESSAGE_ERROR, \
				     GTK_BUTTONS_CLOSE, \
				     msg, ##args); \
    gtk_widget_show (window); \
    gtk_dialog_run (GTK_DIALOG (window)); \
    gtk_widget_destroy (window); \
  } while (0)

/*
 * Menu / Toolbar callbacks.
 */

static void
cb_destroy (GtkWidget * widget, gpointer data)
{
  GstRecCollection *coll = GST_REC_COLLECTION (data);

  gst_object_unref (GST_OBJECT (coll->rec));
  g_object_unref (coll->client);
  g_free (coll);
  gtk_main_quit ();
}

static void
cb_preferences_destroy (GtkWidget * widget, gpointer data)
{
  *(GstRecPreferences **) data = NULL;
}

static void
cb_preferences (GtkWidget * widget, gpointer data)
{
  GstRecCollection *coll = GST_REC_COLLECTION (data);
  static GtkWidget *prefs = NULL;

  if (prefs != NULL) {
    /* re-focus */
    gtk_window_present (GTK_WINDOW (prefs));

    return;
  }

  prefs = gst_rec_preferences_new (coll->rec, coll->client);
  g_signal_connect (G_OBJECT (prefs), "destroy",
      G_CALLBACK (cb_preferences_destroy), &prefs);
  gtk_widget_show (prefs);
}

static void
cb_toggle_toolbar (GtkWidget * widget, gpointer data)
{
  GstRecCollection *coll = GST_REC_COLLECTION (data);

  gconf_client_set_bool (coll->client, GST_REC_KEY_TOOLBAR_VISIBILITY,
      GTK_CHECK_MENU_ITEM (widget)->active, NULL);
}

static void
cb_toggle_video (GtkWidget * widget, gpointer data)
{
  GstRecCollection *coll = GST_REC_COLLECTION (data);

  gconf_client_set_bool (coll->client, GST_REC_KEY_VIDEO_VISIBILITY,
      GTK_CHECK_MENU_ITEM (widget)->active, NULL);
}

static void
cb_toggle_statistics (GtkWidget * widget, gpointer data)
{
  GstRecCollection *coll = GST_REC_COLLECTION (data);

  gconf_client_set_bool (coll->client, GST_REC_KEY_STATISTICS_VISIBILITY,
      GTK_CHECK_MENU_ITEM (widget)->active, NULL);
}

static void
cb_toggle_statusbar (GtkWidget * widget, gpointer data)
{
  GstRecCollection *coll = GST_REC_COLLECTION (data);

  gconf_client_set_bool (coll->client, GST_REC_KEY_STATUSBAR_VISIBILITY,
      GTK_CHECK_MENU_ITEM (widget)->active, NULL);
}

static void
cb_about (GtkWidget * widget, gpointer data)
{
  GtkWidget *about;
  const gchar *authors[] = { "Ronald Bultje <rbultje@ronald.bitfreak.net>",
    NULL
  };

  about = gnome_about_new (PACKAGE_NAME, VERSION,
      "(c) 2003-2005 Ronald Bultje",
      _("A GNOME/GStreamer video capture application"),
      authors, NULL, NULL, NULL);

  gtk_widget_show (about);
}

static void
cb_video_source_changed (GstRec * rec, GstElement * video_source, gpointer data)
{
  GstRecCollection *coll = GST_REC_COLLECTION (data);

  gst_rec_tv_set_element (coll->tv, video_source);
}

static GnomeVFSFileSize
get_free_space (GstRecCollection * coll)
{
  GnomeVFSFileSize size;
  GnomeVFSURI *parent;
  gchar *dir;

  if (coll->file) {
    if ((dir = gnome_vfs_uri_extract_dirname (coll->file))) {
      if ((parent = gnome_vfs_uri_new (dir))) {
        if (gnome_vfs_get_volume_free_space (parent, &size) != GNOME_VFS_OK)
          size = -1;
        gnome_vfs_uri_unref (parent);
      }
      g_free (dir);
    }
  }

  return size;
}

static void
cb_statistics (GstRec * rec,
    GstClockTime time,
    guint64 frames,
    guint64 samples,
    guint lost_frames, guint ins_frames, guint del_frames, gpointer data)
{
  GstRecCollection *coll = GST_REC_COLLECTION (data);

  gst_rec_stats_set (coll->stats, time, frames, samples,
      lost_frames, ins_frames, del_frames, get_free_space (coll));
}

static void
cb_file_changed (GtkWidget * widget, gpointer data)
{
  GstRecCollection *coll = GST_REC_COLLECTION (data);
  GnomeVFSURI *old_uri = coll->file, *new_uri = NULL;
  gchar *filename = gnome_file_entry_get_full_path (coll->fileentry,
      FALSE);
  GnomeVFSFileSize size;

  /* Note that we unref the old URI after creating the new one, because
   * that might prevent a whole tree of unrefs for each dir which will
   * be recreated right thereafter by the new URI. */
  if (filename)
    new_uri = gnome_vfs_uri_new (filename);
  GST_REC_DEBUG ("Filename changed to %s", filename);
  g_object_set (G_OBJECT (coll->output), "uri", new_uri, NULL);
  if (old_uri)
    gnome_vfs_uri_unref (old_uri);
  coll->file = new_uri;

  size = get_free_space (coll);

  if (size != (GnomeVFSFileSize) - 1)
    gst_rec_stats_file (coll->stats, size);
  else
    gst_rec_stats_reset (coll->stats);
}

static gboolean
cb_allow_overwrite (GstElement * element, GnomeVFSURI * uri, gpointer data)
{
  GstRecCollection *coll = GST_REC_COLLECTION (data);
  GtkWidget *window;
  gint res;
  gchar *filename = gnome_vfs_uri_to_string (uri, GNOME_VFS_URI_HIDE_PASSWORD);

  window = gtk_message_dialog_new (GTK_WINDOW (coll->window),
      GTK_DIALOG_MODAL, GTK_MESSAGE_QUESTION, GTK_BUTTONS_YES_NO,
      "The file %s already exists. Overwrite?", filename);
  g_free (filename);
  gtk_widget_show (window);
  res = gtk_dialog_run (GTK_DIALOG (window));
  gtk_widget_destroy (window);

  return (res == GTK_RESPONSE_YES);
}

typedef struct _GstRecErrorHandler
{
  GstRecCollection *coll;
  GstElement *source;
  gchar *msg;
}
GstRecErrorHandler;

static gboolean
cb_error_real (gpointer data)
{
  GstRecErrorHandler *handler = data;
  GstRecCollection *coll = handler->coll;
  GstElementFactory *factory = gst_element_get_factory (handler->source);

  gst_rec_popup_error (coll,
      _("Element %s gave an error:\n\n%s"),
      gst_element_factory_get_longname (factory), handler->msg);
  g_free (handler->msg);
  gst_object_unref (GST_OBJECT (handler->source));
  g_free (handler);

  /* once */
  return FALSE;
}

static void
cb_error (GstElement * element,
    GstElement * originator, GError * error, gchar * debug, gpointer data)
{
  GstRecErrorHandler *handler = g_new (GstRecErrorHandler, 1);

  /* move to our own thread */
  handler->coll = GST_REC_COLLECTION (data);
  handler->source = originator;
  gst_object_ref (GST_OBJECT (originator));
  handler->msg = g_strdup (error->message);
  g_idle_add ((GSourceFunc) cb_error_real, handler);
}

static void
cb_change_state_record (GtkWidget * widget, gpointer data)
{
  GstRecCollection *coll = GST_REC_COLLECTION (data);

  if (gst_rec_set_state (coll->rec, GST_STATE_PLAYING) == GST_STATE_FAILURE) {
    gst_rec_set_state (coll->rec, GST_STATE_NULL);
    gst_rec_popup_error (coll, _("Failed to start recording!"));
  }
}

#if 0
static void
cb_change_state_pause (GtkWidget * widget, gpointer data)
{
  GstRecCollection *coll = GST_REC_COLLECTION (data);

  if (gst_rec_set_state (coll->rec, GST_STATE_PAUSED) == GST_STATE_FAILURE) {
    gst_rec_set_state (coll->rec, GST_STATE_NULL);
    gst_rec_popup_error (coll, _("Failed to pause recording!"));
  }
}
#endif

static void
cb_change_state_stop (GtkWidget * widget, gpointer data)
{
  GstRecCollection *coll = GST_REC_COLLECTION (data);

  if (gst_rec_set_state (coll->rec, GST_STATE_NULL) == GST_STATE_FAILURE) {
    gst_rec_popup_error (coll, _("Failed to stop recording!"));
  }
}

static void
cb_play (GtkWidget * widget, gpointer data)
{
  GstRecCollection *coll = GST_REC_COLLECTION (data);
  GError *error = NULL;
  gchar *cmd, *file;

  file = gnome_vfs_uri_to_string (coll->file, GNOME_VFS_URI_HIDE_NONE);
  cmd = g_strdup_printf ("totem %s", file);
  g_free (file);
  if (!g_spawn_command_line_async (cmd, &error)) {
    g_free (cmd);
    cmd = g_strdup_printf (_("Failed to start totem: %s"), error->message);
    gst_rec_popup_error (coll, cmd);
    g_error_free (error);
  }
  g_free (cmd);
}

static void
cb_sliders (GtkWidget * widget, gpointer data)
{
  GstRecCollection *coll = GST_REC_COLLECTION (data);
  GtkWidget *window;

  /* NULL is a valid response here! It simply means that no
   * single "slider-able" interface is implemented. */
  window = gst_rec_sliders_new (gst_rec_get_video_source (coll->rec),
      gst_rec_get_audio_source (coll->rec), TRUE);
  if (window) {
    gtk_widget_realize (window);
    gst_rec_sliders_attach (GST_REC_SLIDERS (window), widget);
    gtk_widget_show (window);
    gst_rec_sliders_run (GST_REC_SLIDERS (window));
  }
}

/*
 * Menus.
 */

static GnomeUIInfo file_menu[] = {
  GNOMEUIINFO_MENU_EXIT_ITEM (cb_destroy, NULL),
  GNOMEUIINFO_END
};

static GnomeUIInfo view_menu[] = {
  GNOMEUIINFO_TOGGLEITEM (N_("_Toolbar"),
      N_("Toggle visibility of the toolbar"),
      cb_toggle_toolbar, NULL),
  GNOMEUIINFO_TOGGLEITEM (N_("_Video"),
      N_("Toggle visibility of video"),
      cb_toggle_video, NULL),
  GNOMEUIINFO_TOGGLEITEM (N_("S_tatistics"),
      N_("Toggle visibility of statistics"),
      cb_toggle_statistics, NULL),
  GNOMEUIINFO_TOGGLEITEM (N_("_Statusbar"),
      N_("Toggle visibility of the statusbar"),
      cb_toggle_statusbar, NULL),
  GNOMEUIINFO_END
};

#define GNOMEUIINFO_ITEM_QUICKKEY(label, tip, cb, stock_id, quickkey) \
        { GNOME_APP_UI_ITEM, label, tip, (gpointer) cb, \
	  NULL, NULL, GNOME_APP_PIXMAP_STOCK, stock_id, \
	  quickkey, GDK_CONTROL_MASK, NULL }

static GnomeUIInfo actions_menu[] = {
  GNOMEUIINFO_ITEM_QUICKKEY (N_("S_top"), N_("Stop Recording"),
      cb_change_state_stop,
      GST_REC_STOCK_STOP, 't'),
#if 0
  GNOMEUIINFO_ITEM_QUICKKEY (N_("_Pause"), N_("Pause Recording"),
      cb_change_state_pause,
      GST_REC_STOCK_PAUSE, 'p'),
#endif
  GNOMEUIINFO_ITEM_QUICKKEY (N_("_Record"), N_("Start Recording"),
      cb_change_state_record,
      GST_REC_STOCK_RECORD, 'r'),
  GNOMEUIINFO_ITEM_QUICKKEY (N_("_Play"), N_("Play Movie"),
      cb_play,
      GST_REC_STOCK_PLAY, 'p'),
  GNOMEUIINFO_END
};

static GnomeUIInfo settings_menu[] = {
  GNOMEUIINFO_MENU_PREFERENCES_ITEM (cb_preferences, NULL),
  GNOMEUIINFO_END
};

static GnomeUIInfo help_menu[] = {
  GNOMEUIINFO_HELP (PACKAGE),
  GNOMEUIINFO_MENU_ABOUT_ITEM (cb_about, NULL),
  GNOMEUIINFO_END
};

#define GNOMEUIINFO_MENU_ACTIONS_TREE(tree) \
	{ GNOME_APP_UI_SUBTREE_STOCK, N_("_Actions"), \
	  NULL, tree, NULL, NULL, (GnomeUIPixmapType) 0, \
	  NULL, 0, (GdkModifierType) 0, NULL }

static GnomeUIInfo main_menu[] = {
  GNOMEUIINFO_MENU_FILE_TREE (file_menu),
  GNOMEUIINFO_MENU_VIEW_TREE (view_menu),
  GNOMEUIINFO_MENU_ACTIONS_TREE (actions_menu),
  GNOMEUIINFO_MENU_SETTINGS_TREE (settings_menu),
  GNOMEUIINFO_MENU_HELP_TREE (help_menu),
  GNOMEUIINFO_END
};

/*
 * Toolbar.
 */

static GnomeUIInfo toolbar[] = {
  GNOMEUIINFO_ITEM_STOCK (N_("Stop"), N_("Stop Recording"),
      cb_change_state_stop, GST_REC_STOCK_STOP),
#if 0
  GNOMEUIINFO_ITEM_STOCK (N_("Pause"), N_("Pause Recording"),
      cb_change_state_pause, GST_REC_STOCK_PAUSE),
#endif
  GNOMEUIINFO_ITEM_STOCK (N_("Record"), N_("Start Recording"),
      cb_change_state_record, GST_REC_STOCK_RECORD),
  GNOMEUIINFO_ITEM_STOCK (N_("Play"), N_("Play Movie"),
      cb_play, GST_REC_STOCK_PLAY),
  GNOMEUIINFO_SEPARATOR,
  GNOMEUIINFO_ITEM_STOCK (N_("Balance"), N_("Video Color and Audio Volume"),
      cb_sliders, GST_REC_STOCK_SLIDER),
  GNOMEUIINFO_END
};

static void
cb_state_change (GstElement * element,
    GstElementState old_state, GstElementState new_state, gpointer data)
{
  GstRecCollection *coll = GST_REC_COLLECTION (data);

  switch (old_state << 8 | new_state) {
    case GST_STATE_NULL_TO_READY:
      gtk_widget_set_sensitive (GTK_WIDGET (coll->fileentry), FALSE);
      break;
    case GST_STATE_READY_TO_NULL:
      gtk_widget_set_sensitive (GTK_WIDGET (coll->fileentry), TRUE);
      break;
    case GST_STATE_PAUSED_TO_PLAYING:
      gtk_widget_hide (toolbar[1].widget);
      gtk_widget_show (toolbar[0].widget);
      gtk_widget_hide (actions_menu[1].widget);
      gtk_widget_show (actions_menu[0].widget);
      break;
    case GST_STATE_PLAYING_TO_PAUSED:
      gtk_widget_hide (toolbar[0].widget);
      gtk_widget_show (toolbar[1].widget);
      gtk_widget_hide (actions_menu[0].widget);
      gtk_widget_show (actions_menu[1].widget);
      break;
    default:
      break;
  }
}

/*
 * GConf callbacks.
 */

static void
cb_toggle_widget (GConfClient * client,
    guint connection_id, GConfEntry * entry, gpointer data)
{
  GstRecCollection *coll = GST_REC_COLLECTION (data);
  GtkWidget *widget = NULL;
  GnomeUIInfo *menu = NULL;
  gboolean value;

  /* booleans only here */
  g_return_if_fail (entry->value->type == GCONF_VALUE_BOOL);
  value = gconf_value_get_bool (entry->value);

  if (!strcmp (entry->key, GST_REC_KEY_TOOLBAR_VISIBILITY)) {
    BonoboDockItem *item = gnome_app_get_dock_item_by_name (coll->window,
        GNOME_APP_TOOLBAR_NAME);

    widget = GTK_WIDGET (item);

    menu = &view_menu[0];
  } else if (!strcmp (entry->key, GST_REC_KEY_VIDEO_VISIBILITY)) {
    widget = GTK_WIDGET (coll->tv);

    menu = &view_menu[1];
  } else if (!strcmp (entry->key, GST_REC_KEY_STATISTICS_VISIBILITY)) {
    gst_rec_stats_expand (coll->stats, value);

    menu = &view_menu[2];
  } else if (!strcmp (entry->key, GST_REC_KEY_STATUSBAR_VISIBILITY)) {
    widget = coll->window->statusbar;

    menu = &view_menu[3];
  } else {
    g_print ("Unknown key %s\n", entry->key);
  }

  if (!menu)
    return;

  gtk_check_menu_item_set_active (GTK_CHECK_MENU_ITEM (menu->widget), value);

  if (!widget)
    return;

  if (value) {
    gtk_widget_show (widget);
  } else {
    gtk_widget_hide (widget);
  }
}

static void
cb_toggle_element (GConfClient * client,
    guint connection_id, GConfEntry * entry, gpointer data)
{
  GstRecCollection *coll = GST_REC_COLLECTION (data);

  if (!strcmp (entry->key, GST_REC_KEY_VIDEO_SOURCE) ||
      !strcmp (entry->key, GST_REC_KEY_AUDIO_SOURCE) ||
      !strcmp (entry->key, GST_REC_KEY_VIDEO_ENCODER) ||
      !strcmp (entry->key, GST_REC_KEY_AUDIO_ENCODER) ||
      !strcmp (entry->key, GST_REC_KEY_VIDEO_FILTER) ||
      !strcmp (entry->key, GST_REC_KEY_AUDIO_FILTER) ||
      !strcmp (entry->key, GST_REC_KEY_VIDEO_NORM) ||
      !strcmp (entry->key, GST_REC_KEY_VIDEO_INPUT) ||
      !strcmp (entry->key, GST_REC_KEY_AUDIO_INPUT) ||
      !strcmp (entry->key, GST_REC_KEY_MUXER)) {
    const gchar *value;

    /* strings only here */
    g_return_if_fail (entry->value->type == GCONF_VALUE_STRING);
    value = gconf_value_get_string (entry->value);

    if (!strcmp (entry->key, GST_REC_KEY_VIDEO_SOURCE)) {
      gst_rec_set_video_source (coll->rec,
          gst_rec_config_to_video_source (value));
    } else if (!strcmp (entry->key, GST_REC_KEY_AUDIO_SOURCE)) {
      gst_rec_set_audio_source (coll->rec,
          gst_rec_config_to_audio_source (value));
    } else if (!strcmp (entry->key, GST_REC_KEY_VIDEO_ENCODER)) {
      gst_rec_set_video_encoder (coll->rec,
          gst_rec_config_to_video_encoder (value));
    } else if (!strcmp (entry->key, GST_REC_KEY_AUDIO_ENCODER)) {
      gst_rec_set_audio_encoder (coll->rec,
          gst_rec_config_to_audio_encoder (value));
    } else if (!strcmp (entry->key, GST_REC_KEY_MUXER)) {
      gst_rec_set_muxer (coll->rec, gst_rec_config_to_muxer (value));
    } else if (!strcmp (entry->key, GST_REC_KEY_VIDEO_FILTER)) {
      gst_rec_set_video_filter (coll->rec,
          gst_rec_config_to_video_filter (value));
    } else if (!strcmp (entry->key, GST_REC_KEY_AUDIO_FILTER)) {
      gst_rec_set_audio_filter (coll->rec,
          gst_rec_config_to_audio_filter (value));
    } else if (!strcmp (entry->key, GST_REC_KEY_VIDEO_NORM)) {
      const GList *norms;
      GstTuner *tuner;
      GstElement *video_source;
      GstTunerNorm *norm = NULL;

      video_source = gst_rec_get_video_source (coll->rec);

      g_return_if_fail (video_source != NULL);
      g_return_if_fail (GST_IS_TUNER (video_source));

      tuner = GST_TUNER (video_source);
      norms = gst_tuner_list_norms (tuner);

      while (norms) {
        if (!strcasecmp (GST_TUNER_NORM (norms->data)->label, value)) {
          norm = norms->data;
          break;
        }

        norms = norms->next;
      }

      g_return_if_fail (norm != NULL);

      gst_tuner_set_norm (tuner, norm);
    } else if (!strcmp (entry->key, GST_REC_KEY_VIDEO_INPUT)) {
      const GList *inputs;
      GstTuner *tuner;
      GstElement *video_source;
      GstTunerChannel *input = NULL;

      video_source = gst_rec_get_video_source (coll->rec);

      g_return_if_fail (video_source != NULL);
      g_return_if_fail (GST_IS_TUNER (video_source));

      tuner = GST_TUNER (video_source);
      inputs = gst_tuner_list_channels (tuner);

      while (inputs) {
        if (!strcasecmp (GST_TUNER_CHANNEL (inputs->data)->label, value) &&
            GST_TUNER_CHANNEL_HAS_FLAG (GST_TUNER_CHANNEL (inputs->data),
                GST_TUNER_CHANNEL_INPUT)) {
          input = inputs->data;
          break;
        }

        inputs = inputs->next;
      }

      g_return_if_fail (input != NULL);

      gst_tuner_set_channel (tuner, input);
    } else if (!strcmp (entry->key, GST_REC_KEY_AUDIO_INPUT)) {
      const GList *tracks;
      GstMixer *mixer;
      GstMixerTrack *track;
      GstElement *audio_source;

      audio_source = gst_rec_get_audio_source (coll->rec);

      g_return_if_fail (audio_source != NULL);
      g_return_if_fail (GST_IS_MIXER (audio_source));

      mixer = GST_MIXER (audio_source);
      tracks = gst_mixer_list_tracks (mixer);

      while (tracks) {
        gboolean match;

        track = tracks->data;
        match = !strcasecmp (track->label, value);

        gst_mixer_set_mute (mixer, track, !match);
        gst_mixer_set_record (mixer, track, match);

        tracks = tracks->next;
      }
    }
  } else if (!strcmp (entry->key, GST_REC_KEY_VIDEO_WIDTH) ||
      !strcmp (entry->key, GST_REC_KEY_VIDEO_HEIGHT) ||
      !strcmp (entry->key, GST_REC_KEY_AUDIO_SAMPLERATE) ||
      !strcmp (entry->key, GST_REC_KEY_AUDIO_CHANNELS)) {
    int value;

    /* strings only here */
    g_return_if_fail (entry->value->type == GCONF_VALUE_INT);
    value = gconf_value_get_int (entry->value);

    if (!strcmp (entry->key, GST_REC_KEY_VIDEO_WIDTH)) {
      gst_rec_set_video_width (coll->rec, value);
    } else if (!strcmp (entry->key, GST_REC_KEY_VIDEO_HEIGHT)) {
      gst_rec_set_video_height (coll->rec, value);
    } else if (!strcmp (entry->key, GST_REC_KEY_AUDIO_SAMPLERATE)) {
      gst_rec_set_audio_samplerate (coll->rec, value);
    } else if (!strcmp (entry->key, GST_REC_KEY_AUDIO_CHANNELS)) {
      gst_rec_set_audio_channels (coll->rec, value);
    }
  } else {
    g_print ("Unknown key %s\n", entry->key);
    return;
  }
}

/*
 * Load all GConf entries. We do this by re-setting each entry...
 * Also initiate the config when we haven't done this before yet.
 * Why doesn't GConf have a function for this directly?...
 */

static void
load_config (GstRecCollection * coll)
{
  GSList *dirs, *entries;
  const gchar *dir;
  const GConfEntry *entry;
  GConfValue *value, *fake;
  GList *elements;

  /* first-time setup? */
  if (!(value = gconf_client_get (coll->client, GST_REC_KEY_SETUP, NULL)) ||
      value->type != GCONF_VALUE_BOOL || !gconf_value_get_bool (value)) {
    GtkWidget *druid;

    druid = gst_rec_druid_new (coll->rec, coll->client);
    gtk_window_set_transient_for (GTK_WINDOW (druid),
        GTK_WINDOW (coll->window));
    gtk_window_set_modal (GTK_WINDOW (druid), TRUE);
    gtk_widget_show (druid);

    return;
  }

  /* Now, before we load everything, we first want to do all device
   * probing. Why? Simple! If we load OSS (or any) and we add it, it
   * will be set to READY. Probing can only be done in NULL, so any
   * try to probe devices will fail. */
  elements = gst_rec_list_video_sources ();
  gst_rec_list_replace (&elements, NULL);
  elements = gst_rec_list_audio_sources ();
  gst_rec_list_replace (&elements, NULL);

  /* bleh... GConf, implement this sanely! */
  dirs = gconf_client_all_dirs (coll->client, GST_REC_KEY_BASEDIR, NULL);
  while (dirs) {
    dir = dirs->data;

    entries = gconf_client_all_entries (coll->client, dir, NULL);
    while (entries) {
      if (!(entry = entries->data) || !entry->value)
        goto next_entry;
      if (!(value = gconf_value_copy (entry->value)))
        goto next_entry;

      if (value->type == GCONF_VALUE_BOOL) {
        fake = gconf_value_new (GCONF_VALUE_STRING);
        gconf_value_set_string (fake, "bla");
      } else {
        fake = gconf_value_new (GCONF_VALUE_BOOL);
        gconf_value_set_bool (fake, FALSE);
      }
      gconf_client_set (coll->client, entry->key, fake, NULL);
      gconf_client_set (coll->client, entry->key, value, NULL);

    next_entry:
      entries = entries->next;
    }

    dirs = dirs->next;
  }
}

/*
 * Window creation.
 */

void
create_window (void)
{
  GstRecCollection *coll = g_new0 (GstRecCollection, 1);
  GtkWidget *window, *appbar, *area, *fileentry, *stats, *box, *tv;
  GstRec *rec;
  GstElement *filesink;

  /* gconf */
  coll->client = gconf_client_get_default ();
  gconf_client_add_dir (coll->client, GST_REC_KEY_BASEDIR,
      GCONF_CLIENT_PRELOAD_RECURSIVE, NULL);

  /* UI config entries */
  gconf_client_notify_add (coll->client, GST_REC_KEY_UIDIR,
      cb_toggle_widget, coll, NULL, NULL);

  /* Gst config entries */
  gconf_client_notify_add (coll->client, GST_REC_KEY_GSTDIR,
      cb_toggle_element, coll, NULL, NULL);

  /* backend setup */
  rec = gst_rec_new ();
  coll->rec = rec;
  g_signal_connect (G_OBJECT (rec), "state_change",
      G_CALLBACK (cb_state_change), coll);
  g_signal_connect (G_OBJECT (rec), "error", G_CALLBACK (cb_error), coll);
  filesink = gst_element_factory_make ("gnomevfssink", "output-0");
  g_signal_connect (filesink, "allow-overwrite",
      G_CALLBACK (cb_allow_overwrite), coll);
  gst_rec_set_output (rec, filesink);
  coll->output = filesink;

  /* main window */
  window = gnome_app_new (PACKAGE, PACKAGE_NAME);
  coll->window = GNOME_APP (window);

  /* menu bar */
  gnome_app_create_menus_with_data (GNOME_APP (window),
      main_menu, (gpointer) coll);
  GTK_CHECK_MENU_ITEM (view_menu[0].widget)->active = TRUE;
  GTK_CHECK_MENU_ITEM (view_menu[1].widget)->active = TRUE;
  GTK_CHECK_MENU_ITEM (view_menu[2].widget)->active = TRUE;
  GTK_CHECK_MENU_ITEM (view_menu[3].widget)->active = TRUE;

  /* toolbar */
  gnome_app_create_toolbar_with_data (GNOME_APP (window), toolbar,
      (gpointer) coll);

  /* main area */
  area = gtk_vbox_new (FALSE, 2);
  gtk_container_set_border_width (GTK_CONTAINER (area), 2);
  gnome_app_set_contents (GNOME_APP (window), area);

  /* TV widget */
  tv = gst_rec_tv_new (NULL);
  gtk_box_pack_start (GTK_BOX (area), tv, TRUE, TRUE, 0);
  g_signal_connect (G_OBJECT (rec), "video_source_changed",
      G_CALLBACK (cb_video_source_changed), coll);
  coll->tv = GST_REC_TV (tv);

  /* a second box containing the rest (different padding) */
  box = gtk_vbox_new (FALSE, 6);
  gtk_container_set_border_width (GTK_CONTAINER (box), 4);
  gtk_box_pack_start (GTK_BOX (area), box, FALSE, FALSE, 0);

  /* file selection widget */
  fileentry = gnome_file_entry_new ("file-selection",
      _(PACKAGE_NAME ": Select target file"));
  g_signal_connect (G_OBJECT (fileentry), "changed",
      G_CALLBACK (cb_file_changed), coll);
  coll->fileentry = GNOME_FILE_ENTRY (fileentry);
  gtk_box_pack_start (GTK_BOX (box), fileentry, TRUE, TRUE, 0);

  /* statistics */
  stats = gst_rec_stats_new (TRUE);
  g_signal_connect (G_OBJECT (rec), "statistics",
      G_CALLBACK (cb_statistics), coll);
  coll->stats = GST_REC_STATS (stats);
  gtk_table_set_col_spacings (GTK_TABLE (stats), 6);
  gtk_box_pack_start (GTK_BOX (box), stats, TRUE, TRUE, 0);

  /* status bar */
  appbar = gnome_appbar_new (FALSE, TRUE, GNOME_PREFERENCES_USER);
  gnome_app_set_statusbar (GNOME_APP (window), appbar);
  gnome_app_install_appbar_menu_hints (GNOME_APPBAR (appbar), main_menu);

  /* exit signal */
  g_signal_connect (window, "destroy", G_CALLBACK (cb_destroy),
      (gpointer) coll);

  /* show off */
  gtk_widget_show_all (window);
  gtk_widget_hide (toolbar[0].widget);
  gtk_widget_hide (actions_menu[0].widget);
  load_config (coll);
}
